use clap::{Parser, Subcommand};
use image::{imageops, DynamicImage, GenericImageView, ImageReader, Pixel};
use termion::{color, terminal_size};

trait TextImageRenderer {
    fn render(&self, image: DynamicImage, width: u32, height: u32) -> String;
}

struct UnicodeBlockRendererConfig {
    alpha_threshold: u8,
    block_char: String,
}

struct UnicodeBlockRenderer {
    config: UnicodeBlockRendererConfig,
}
impl TextImageRenderer for UnicodeBlockRenderer {
    fn render(&self, image: DynamicImage, width: u32, height: u32) -> String {
        let scaled_image = image.resize(width, height, imageops::FilterType::Gaussian);

        // This might not match `(environment.width, environment.height)` because resize preserves
        // aspect ratio on the scaled image
        let (scaled_x, scaled_y) = scaled_image.dimensions();

        let mut out = String::new();
        for y in 0..scaled_y {
            for x in 0..scaled_x {
                let pix = scaled_image.get_pixel(x.into(), y.into()).to_rgba();
                if pix[3] >= self.config.alpha_threshold {
                    out.push_str(&format!(
                        "{}{}",
                        color::Fg(color::Rgb(pix[0], pix[1], pix[2])),
                        self.config.block_char,
                    ));
                } else {
                    out.push_str(&format!("{} ", color::Fg(color::Reset)));
                }
            }
            out.push_str("\n");
        }

        out
    }
}

#[derive(Parser, Debug)]
#[command(name = "picsl")]
#[command(version = "0.1")]
#[command(about = "Render images in the terminal", long_about = None)]
struct Cli {
    /// Path to image
    #[arg(short, long)]
    image: String,

    #[command(subcommand)]
    renderer: RendererCommands,
}

#[derive(Subcommand, Debug)]
enum RendererCommands {
    UnicodeBlock {
        /// A pixel's alpha channel must be higher than this threshold to be rendered.
        #[arg(long, default_value_t = 0)]
        alpha_threshold: u8,

        /// The (potentially unicode) character to use for filled-in values
        #[arg(short, long, default_value_t = String::from("█"))]
        block_char: String,
    },
}

fn get_renderer(renderer_command: RendererCommands) -> impl TextImageRenderer {
    match renderer_command {
        RendererCommands::UnicodeBlock {
            alpha_threshold,
            block_char,
        } => UnicodeBlockRenderer {
            config: UnicodeBlockRendererConfig {
                alpha_threshold,
                block_char,
            },
        },
    }
}

fn main() -> Result<(), Box<dyn std::error::Error>> {
    let args = Cli::parse();
    let image_path = args.image;

    // Load the image and resize it to terminal size
    let (term_x, term_y) = terminal_size().expect("Terminal size");
    let image_buf = ImageReader::open(image_path)?.decode()?;
    let renderer = get_renderer(args.renderer);
    let rendered_image_text = renderer.render(image_buf, term_x.into(), term_y.into());

    print!(
        "{}{}{}{}{}",
        // Blank canvas
        termion::clear::All,
        // Go to top
        termion::cursor::Goto(1, 1),
        // Display image
        rendered_image_text,
        // Reset colors in case we do something else later
        color::Fg(color::Reset),
        color::Bg(color::Reset),
    );

    Ok(())
}
